import Sequelize = require("sequelize");
const Op = Sequelize.Op;

export enum CONNECTING_OPERATORS {
  OR = "OR",
  AND = "AND"
}

export enum OPERATORS {
  EQ = "EQ",
  NOT = "NOT",
  IN = "IN",
  NOTIN = "NOTIN",
  BETWEEN = "BETWEEN",
  NOTBETWEEN = "NOTBETWEEN",
  LT = "LT",
  LTE = "LTE",
  GT = "GT",
  GTE = "GTE",
  NULL = "NULL",
  LIKE = "LIKE",
  NOTLIKE = "NOTLIKE"
}

const SEQUELIZE_OP_MAP: {
  [key: string]: any;
} = {
  [CONNECTING_OPERATORS.OR]: Op.or,
  [CONNECTING_OPERATORS.AND]: Op.and,
  [OPERATORS.EQ]: Op.eq,
  [OPERATORS.NOT]: Op.not,
  [OPERATORS.IN]: Op.in,
  [OPERATORS.NOTIN]: Op.notIn,
  [OPERATORS.BETWEEN]: Op.between,
  [OPERATORS.NOTBETWEEN]: Op.notBetween,
  [OPERATORS.LT]: Op.lt,
  [OPERATORS.LTE]: Op.lte,
  [OPERATORS.GT]: Op.gt,
  [OPERATORS.GTE]: Op.gte,
  [OPERATORS.NULL]: Op.is,
  [OPERATORS.LIKE]: Op.like,
  [OPERATORS.NOTLIKE]: Op.notLike
};

const OPERTAOR_VALUE_TRANSFORMER: {
  [key: string]: any;
} = {
  [OPERATORS.LIKE]: (val: string) => `%${val}%`,
  [OPERATORS.NOTLIKE]: (val: string) => `%${val}%`,
  [OPERATORS.IN]: (val: string) => val.split("-."),
  [OPERATORS.NOTIN]: (val: string) => val.split("-."),
  [OPERATORS.BETWEEN]: (val: string | string[]) =>
    Array.isArray(val) ? val : val.split(",")
};

module.exports.SequelizeQueryStringParser = (req: any, res: any, next: any) => {
  const {
    filter,
    order,
    page: queryPage = 1,
    perPage,
    includes_filter,
    includes
  } = req.query;

  const page = queryPage > 0 ? queryPage : 1;

  /**
   * Parse query string into sequelize readable objects
   */
  const sFilter = filter ? parseFilter(filter) : null;
  const sOrder = order ? parseOrder(order) : null;

  let parsedIncludes = includes
    ? includes
        .split(",")
        .map((association: string) => parseIncludes(association))
    : null;

  if (includes_filter) {
    Object.keys(includes_filter).forEach(key => {
      parsedIncludes = parsedIncludes.map(
        (association: { association: string }) => {
          if (key === association.association) {
            return {
              ...association,
              where: parseFilter(includes_filter[key])
            };
          } else {
            return association;
          }
        }
      );
    });
  }

  /**
   * Set the sequelize query object to the request so that it can be used later
   */
  req.sequelizeOptions = {
    distinct: parsedIncludes && parsedIncludes.length > 0,
    include: parsedIncludes,
    limit: perPage ? perPage : null,
    offset: perPage && page ? perPage * (page - 1) : 0,
    order: [sOrder],
    where: sFilter
  };

  next();
};

const parseFilter = (filterObject: {
  op: OPERATORS | CONNECTING_OPERATORS;
  key: string;
  value: any;
}): any => {
  const { op = null, key = null, value = null } = filterObject;

  if (op === null) {
    return null;
  }

  if ((Object.values(CONNECTING_OPERATORS) as any).includes(op)) {
    return {
      [SEQUELIZE_OP_MAP[op]]: value.map((val: any) => ({
        ...parseFilter(val)
      }))
    };
  }

  if ((Object.values(OPERATORS) as any).includes(op) && key !== null) {
    return {
      [key]: {
        [SEQUELIZE_OP_MAP[op]]: OPERTAOR_VALUE_TRANSFORMER[op]
          ? OPERTAOR_VALUE_TRANSFORMER[op](value)
          : value
      }
    };
  }

  return null;
};

const parseIncludes = (includesText: string): any => {
  const associationNames = includesText.split(/\.(.+)/);

  return {
    association: associationNames[0].replace("*", ""),
    ...(associationNames.length > 1
      ? { include: parseIncludes(associationNames[1]) }
      : {}),
    required: associationNames[0].includes("*")
  };
};

const parseOrder = (orderString: string) => {
  return orderString.split(" ");
};

module.exports.parseFilter = parseFilter;
module.exports.parseOrder = parseOrder;
